# models.py

from typing import List, Optional, Union
from pydantic import BaseModel, Field
from datetime import date
from rich.console import Console
from rich.table import Table


class DataValueInformation(BaseModel):
    name: str = Field(
        ...,
        description="The name of the input or output, must be unique,\
will be used to identify the input or output",
        example="my_data_01",
    )
    type: str = Field(
        ...,
        description="The type of the input or output, only string, number,\
boolean and array are supported",
        example="string",
        enum=["string", "number", "boolean", "array"],
    )
    description: Optional[str] = Field(
        None,
        description="The description of the input or output",
        example="This data is used to do something",
    )


class ValueInformationString(DataValueInformation):
    type: str = Field("string", Literal=True, example="string", enum=["string"])
    availableValues: Optional[List[str]] = Field(
        None,
        description="The list of available values for this input",
        example=["my value", "my other value"],
    )
    default: Optional[str] = Field(
        None, description="The default value for this input", example="my value"
    )


class ValueInformationNumber(DataValueInformation):
    type: str = Field("number", Literal=True, example="number", enum=["number"])
    availableValues: Optional[List[float]] = Field(
        None,
        description="The list of available values for this input",
        example=[1, 2, 3],
    )
    min: Optional[float] = Field(
        None, description="The minimum value for this input", example=0
    )
    max: Optional[float] = Field(
        None, description="The maximum value for this input", example=10
    )
    default: Optional[float] = Field(
        None, description="The default value for this input", example=5
    )


class ValueInformationBoolean(DataValueInformation):
    type: str = Field("boolean", Literal=True, example="boolean", enum=["boolean"])
    default: Optional[bool] = Field(
        None, description="The default value for this input", example=True
    )


class ValueInformationArray(DataValueInformation):
    type: str = Field("array", Literal=True, example="array", enum=["array"])
    arrayType: str = Field(
        ...,
        description="Specify type of the array, only string, number and\
boolean supported",
        example="string",
        enum=["string", "number", "boolean", "dict", "array"],
    )
    lengthMin: Optional[int] = Field(
        None, description="The minimum length of the array", example=0
    )
    lengthMax: Optional[int] = Field(
        None, description="The maximum length of the array", example=10
    )


InputOutputType = Union[
    ValueInformationString,
    ValueInformationNumber,
    ValueInformationBoolean,
    ValueInformationArray,
]


class Algorithm(BaseModel):
    id: str = Field(
        ...,
        description="The id of the algorithm, must be unique, will be used\
to identify the algorithm",
        example="my-algorithm-01",
    )
    name: Optional[str] = Field(
        None, description="The name of the algorithm", example="My algorithm 01"
    )
    description: Optional[str] = Field(
        None,
        description="The description of the algorithm",
        example="This algorithm is used to do something",
    )
    tags: Optional[List[str]] = Field(
        None, description="The list of tags of the algorithm", example=["tag1", "tag2"]
    )
    author: Optional[str] = Field(
        None, description="The author of the algorithm", example="Ada Lovelace"
    )
    creationDate: Optional[date] = Field(
        ...,
        description="The creation date of the algorithm, ISO 8601 format, YYYY-MM-DD",
        example="2023-01-01",
    )
    updateDate: Optional[date] = Field(
        None,
        description="The last algorithm update date, ISO 8601 format, YYYY-MM-DD",
        example="2023-03-20",
    )
    version: Optional[str] = Field(
        ..., description="The version of the algorithm", example="0.1.0"
    )
    inputs: List[InputOutputType] = Field(
        ..., description="The list of inputs of the algorithm"
    )
    outputs: List[InputOutputType] = Field(
        ..., description="The list of outputs of the algorithm"
    )

    def get_table(self):
        # Algorithm Details
        table = Table(width=80)
        table.add_column(
            self.id,
            style="cyan",
            no_wrap=True,
            justify="right",
            width=20,
        )

        table_title = ""
        if self.name:
            table_title += self.name
        if self.version:
            table_title += f" [blue][{self.version}][/blue]"

        table.add_column(table_title, width=60)

        if self.tags:
            tags_text = ""
            for tag in self.tags:
                tags_text += f"[blue]⎣{tag}⎤[/blue] "
            table.add_row("Tags:", tags_text + "\n")
        if self.description:
            table.add_row("Description:", self.description)
        if self.author:
            table.add_row("Author:", self.author)
        if self.creationDate:
            table.add_row("Creation Date:", str(self.creationDate))
        if self.updateDate:
            table.add_row("Update Date:", str(self.updateDate))

        # Inputs
        if self.inputs:
            for input in self.inputs:
                input_text = f"[bold blue]{input.type}[/bold blue]"
                if input.type == "array" and hasattr(input, "arrayType"):
                    input_text += f"[blue][[italic]{input.arrayType}[/italic]][/blue] "
                else:
                    input_text += " "

                input_text += f"[italic]{input.description}[/italic]"

                table.add_row(
                    f"[bold green]{input.name}[/bold green]",
                    input_text,
                )

        # Outputs
        if self.outputs:
            for output in self.outputs:
                table.add_row(
                    f"[bold magenta]{output.name}[/bold magenta]",
                    f"[bold blue]{output.type}[/bold blue] "
                    + f"[italic]{output.description}[/italic]",
                )

        return table

    def print_table(self):
        console = Console()
        console.print(self.get_table())

    def __str__(self):
        from io import StringIO

        # Display the table, capture the output and return it as a string
        console = Console()
        with StringIO() as buffer:
            # Render the table to the buffer
            console.print(self.get_table())
            # Return the buffer contents as a string
            return buffer.getvalue()



from typing import Callable, List, Dict, Optional
from algo_provider.utils.parser import extract_algorithm_metadata
from algo_provider.models.algorithm import Algorithm
from algo_provider.app import start_api_server
from rich.console import Console
from rich.panel import Panel


class AlgoProvider:
    def __init__(self):
        self.algorithms: Dict[str, Algorithm] = {}
        self.algorithms_functions: Dict[str, Callable] = {}

    def add_algo(
        self,
        func: Callable,
        tags: List[str] = [],
        author: Optional[str] = None,
        version: Optional[str] = None,
        creation_date: Optional[str] = None,
        update_date: Optional[str] = None,
    ) -> Algorithm:
        """
        Adds an algorithm to the provider.

        Parameters:
            func (Callable): The function to add as an algorithm.
            tags (List[str]): The tags of the algorithm.
            version (Optional[str]): The version of the algorithm.
            creation_date (Optional[str]): The creation date of the algorithm.
            update_date (Optional[str]): The update date of the algorithm.

        Returns:
            Algorithm: The algorithm object.
        """
        algo_metadata = extract_algorithm_metadata(func)

        algorithm = Algorithm(
            **algo_metadata,
            tags=tags,
            author=author,
            version=version,
            creationDate=creation_date,
            updateDate=update_date,
        )

        self.algorithms[algorithm.id] = algorithm
        self.algorithms_functions[algorithm.id] = func
        return algorithm

    def get_algorithms(self) -> List[Algorithm]:
        return list(self.algorithms.values())

    def run_algorithm(self, algorithm_id: str, inputs: Dict):
        # Verify if the algorithm exists
        if algorithm_id not in self.algorithms:
            raise ValueError(f"Algorithm '{algorithm_id}' not found.")

        # Inputs Example:
        # {
        #     "inputs": [
        #         {"name": "input1", "value": 12},
        #         {"name": "input2", "value": 22},
        #     ],
        # }

        # Transform this to a dictionary {input_name: input_value}
        inputs = {input["name"]: input["value"] for input in inputs["inputs"]}

        # Call the algorithm function
        try:
            function = self.algorithms_functions[algorithm_id]
            result = function(**inputs)
        except Exception as e:
            # Print the error message
            error_message = f"[bold red]Error running algorithm \
 '{algorithm_id}':\n{str(e)}[/bold red]"
            console = Console()
            console.print(
                Panel(
                    error_message,
                    title="[bold red]Error[/bold red]",
                    width=80,
                    border_style="bold",
                    style="red",
                )
            )

            # Raise the error
            raise ValueError(str(e))

        # Return the result
        return [{"name": "result", "value": result}]

    def start_server(self, host="0.0.0.0", port=8000):
        # Print the server information
        console = Console()
        console.print(
            Panel(
                "The Easy Algorithm Provider is being started..."
                + f"\n\n[bold]API Server[/bold]: http://{host}:{port}"
                + f"\n[bold]Number of Algorithms[/bold]: {len(self.get_algorithms())}",
                title="Easy Algorithm Provider",
                width=80,
                border_style="bold",
            )
        )

        # Print the creation message of each algorithm
        for algorithm in self.get_algorithms():
            algorithm.print_table()

        start_api_server(self, host, port)



APP_VERSION = "1.0.1"


def start_api_server(provider, host, port):
    import uvicorn
    from fastapi import FastAPI, HTTPException
    from typing import List, Dict, Any
    from algo_provider.models.algorithm import Algorithm
    from fastapi.middleware.cors import CORSMiddleware

    app = FastAPI(
        title="Algo API",
        version=APP_VERSION,
        description="API for algorithm provider",
    )

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    @app.get("/algorithms", response_model=List[Algorithm])
    def get_algorithms():
        return provider.get_algorithms()

    @app.post("/algorithms/{algorithm_id}/run")
    def run_algorithm(algorithm_id: str, inputs: Dict[str, Any]):
        try:
            return provider.run_algorithm(algorithm_id, inputs)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e))

    uvicorn.run(app, host=host, port=port)



# Template algorithms


def add_floats(number1: float, number2: float) -> float:
    """
    Adds two numbers together and returns the result.

    Args:
        number1 (float): The first number to add.(integer or float).
        number2 (float): The second number to add.(integer or float).

    Returns:
        float: The result of the addition of number1 and number2.
    """

    return number1 + number2


def concatenate_strings(str1: str, str2: str) -> str:
    """
    Concatenates two input strings and returns the result.

    Args:
        str1 (str): The first string to concatenate.
        str2 (str): The second string to concatenate.

    Returns:
        str: The concatenated result of str1 and str2.
    """
    return str1 + str2


def statistics_of_list(data: list) -> list:
    """
    Calculates the mean and median of a list of numbers.

    Args:
        data (list): List of numbers to calculate the statistics on.

    Returns:
        list: The calculated mean and median.
    """
    average = sum(data) / len(data)
    median = data[len(data) // 2]

    return [average, median]


def multiply_lists(list1: list, list2: list) -> list:
    """
    Multiplies two lists element-wise.

    Args:
        list1 (list): The first list to multiply.
        list2 (list): The second list to multiply.

    Returns:
        list: The element-wise multiplication of list1 and list2.
    """
    return [x * y for x, y in zip(list1, list2)]


def moving_average(data: list, periods: int = 3) -> list:
    """Defines the logic of the algorithm

    Args:
        data (list): List of numbers to calculate the moving average on.
        periods (int): Number of periods for the moving average.

    Returns:
        list: The calculated moving average.
    """
    return [
        sum(data[i : i + periods]) / periods for i in range(len(data) - periods + 1)
    ]



from setuptools import setup, find_packages
from algo_provider.app import APP_VERSION

# Import the README and use it as the long description
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Setup the package
setup(
    name="algo-provider",
    version=APP_VERSION,
    description="Start your own Algo provider from a Python function,\
 interface it DebiAI or something else",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/debiai/eaty-algo-provider",
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=[
        "fastapi==0.115.4",
        "uvicorn==0.32.0",
        "rich==13.9.4",
    ],
    entry_points={},
)



import inspect
import re

TYPE_MAPPING = {
    "int": "number",
    "float": "number",
    "str": "string",
    "bool": "boolean",
    "list": "array",
    "Any": "any",
}


def extract_algorithm_metadata(function):
    # Get the docstring of the function (if it exists)
    docstring = function.__doc__

    # Parse function signature using inspect module
    signature = inspect.signature(function)

    # Initialize the structure for the metadata
    algorithm_metadata = {
        "id": function.__name__,
        "name": function.__name__.replace("_", " ").title(),
        "description": (
            docstring.split("\n\n")[0].strip() if docstring else ""
        ),  # Default empty if no docstring
        "inputs": [],
        "outputs": [],
    }

    # Extraction of the function arguments with their types and descriptions
    args_details = {}
    return_details = {"type": None, "description": None}
    if docstring:
        # Docstring example::
        # This is a simple algorithm that adds two numbers together.
        # Parameters:
        #     input1 (int): The first number to add.
        #     input2 (int): The second number to add.
        # Returns:
        #     int: The sum of the two numbers.

        # This regular expression match the parameters in the docstring of the function
        # It looks for a word followed by a colon and the type in parenthesis
        input_pattern = re.compile(r"(\w+)\s*\(([^)]+)\):\s*([^\n]+)")

        for match in input_pattern.finditer(docstring):
            param_name = match.group(1)
            param_type = match.group(2)
            param_description = match.group(3)

            args_details[param_name] = {
                "type": TYPE_MAPPING.get(param_type, param_type),
                "description": param_description,
            }

        # This regular expression match the return type in the docstring of the function
        # It looks for the word "Returns:" followed by the type
        return_pattern = re.compile(r"Returns:\s*([^\n]+)")

        match = return_pattern.search(docstring)
        if match:
            # match.group(1): "int: The sum of the two numbers."
            if match.group(1):
                return_details["type"] = match.group(1).split(":")[0].strip()
                return_details["description"] = match.group(1).split(":")[1].strip()

    # Gets inputs from function signature
    for param_name, param in signature.parameters.items():
        param_data = args_details.get(param_name, {})
        param_type = param_data.get("type")

        if not param_type:
            if param.annotation != inspect.Parameter.empty:
                param_type = TYPE_MAPPING.get(
                    param.annotation.__name__, param.annotation.__name__
                )
            else:
                param_type = "any"

        # Input
        input_metadata = {
            "name": param_name,
            "description": param_data.get("description", ""),
            "type": param_type,
        }

        # Array type
        if param_type == "array" and hasattr(param.annotation, "__args__"):
            array_type = TYPE_MAPPING.get(
                param.annotation.__args__[0].__name__,
                param.annotation.__args__[0].__name__,
            )
            input_metadata["arrayType"] = array_type

        # Add default value
        if param.default != inspect.Parameter.empty:
            input_metadata["default"] = param.default

        # Add to the inputs list
        algorithm_metadata["inputs"].append(input_metadata)

    # Output
    if return_details["type"]:
        return_type = TYPE_MAPPING.get(return_details["type"], return_details["type"])
    else:
        return_type = (
            TYPE_MAPPING.get(str(signature.return_annotation.__name__), "any")
            if signature.return_annotation != inspect.Signature.empty
            else "any"
        )

    if return_details["description"]:
        return_description = return_details["description"]
    else:
        return_description = f"The result of {function.__name__}."

    algorithm_metadata["outputs"].append(
        {
            "name": "result",
            "description": return_description,
            "type": return_type,
        }
    )

    return algorithm_metadata
